<div class="dropdown" id="version_switcher">
    <button type="button" class="btn btn-sm navbar-btn dropdown-toggle" id="version_switcher_button" data-toggle="dropdown">
        {{ current_version.name  }}
        <span class="caret"></span>
    </button>
    <div id="version_switcher_menu" class="dropdown-menu list-group-flush py-0" aria-labelledby="version_switcher_button">
        <dl>
            <dt>Releases</dt>
            {%- for item in versions.tags %}
            <dd><a href="{{ item.url }}">{{ item.name }}</a></dd>
            {%- endfor %}
        </dl>        
        <dl>
            <dt>Branches</dt>
            {%- for item in versions.branches %}
            <dd><a href="{{ item.url }}">{{ item.name }}</a></dd>
            {%- endfor %}
        </dl>        
    </div>
</div>
